#include "Pledge.h"
#include "fs/FSRegistry.h"
#include "fs/VFS.h"
#include "fs/devpts/FileSystem.h"
#include "fs/ext2/FileSystem.h"
#include "fs/tmpfs/FileSystem.h"
#include "memory/MemoryManager.h"
#include "sys/Syscall.h"
#include "thread/Scheduler.h"

Result<u64> sys_mount(Registers*, SyscallArgs args)
{
    auto target = TRY(MemoryManager::strdup_from_user(args[0]));
    auto fstype = TRY(MemoryManager::strdup_from_user(args[1]));
    auto source = TRY(MemoryManager::strdup_from_user(args[2]));

    auto* current = Process::current();
    TRY(check_pledge(current, Promise::p_mount));
    if (current->credentials().euid != 0) return err(EPERM);

    auto get_source = [current, &source]() -> Result<SharedPtr<Device>> {
        auto inode = TRY(VFS::resolve_path(source.chars(), current, current->current_directory));
        if (inode->type() != VFS::InodeType::BlockDevice) return err(ENOTBLK);
        dev_t device_id = inode->metadata().devid;
        return TRY(DeviceRegistry::fetch_special_device(luna_dev_major(device_id), luna_dev_minor(device_id)));
    };

    SharedPtr<VFS::FileSystem> fs;

    auto maybe_virtual = FSRegistry::lookup_virtual_filesystem(fstype.view());
    if (maybe_virtual.has_value())
    {
        auto factory = maybe_virtual.release_value();
        fs = TRY(factory());
    }
    else
    {
        auto factory = TRY(FSRegistry::lookup_physical_filesystem(fstype.view()));
        auto device = TRY(get_source());
        fs = TRY(factory(device));
    }

    TRY(VFS::mount(target.chars(), fs, current, current->current_directory));

    return 0;
}

Result<u64> sys_umount(Registers*, SyscallArgs args)
{
    auto target = TRY(MemoryManager::strdup_from_user(args[0]));

    auto* current = Process::current();
    TRY(check_pledge(current, Promise::p_mount));
    if (current->credentials().euid != 0) return err(EPERM);

    TRY(VFS::umount(target.chars(), current, current->current_directory));

    return 0;
}

Result<u64> sys_pivot_root(Registers*, SyscallArgs args)
{
    auto new_root = TRY(MemoryManager::strdup_from_user(args[0]));
    auto put_old = TRY(MemoryManager::strdup_from_user(args[1]));

    auto* current = Process::current();
    TRY(check_pledge(current, Promise::p_mount));
    if (current->credentials().euid != 0) return err(EPERM);

    TRY(VFS::pivot_root(new_root.chars(), put_old.chars(), current->current_directory));

    return 0;
}
